//
// Copyright 2003 (c) Kevin Atkinson under the GNU GPL license version
// 2.0.  You should have received a copy of the GPL license along with
// this program if you did not you can find it at http://www.gnu.org/.
//

#include <stddef.h>
#include <stdlib.h>
#include <assert.h>

#include <cmath>
#include <algorithm>

using std::swap;
using std::abs;
using std::min;
using std::max;
typedef unsigned char byte;
static inline bool is_odd(int x) {return x & 1;}
static inline bool is_even(int x) {return !is_odd(x);}

#define REGPARM __attribute__ (( regparm(3) ))

#include "avisynth_c.h"

////////////////////////////////////////////////////////////////////
//
//

#define DEBUG

#ifdef DEBUG

#include <ostream>
#include <ext/stdio_filebuf.h>
#include <windows.h>
#include <io.h>
#include <fcntl.h>
using std::ostream;
using std::endl;

struct Log : public ostream {
  __gnu_cxx::stdio_filebuf<char> fbuf;
  int get_fd() {
    AllocConsole();
    HANDLE h = GetStdHandle( STD_OUTPUT_HANDLE );
    return _open_osfhandle( (long) h, _O_TEXT );}
  Log() : ostream(&fbuf), fbuf(get_fd(), std::ios::out, false, 128) {}
} clog;

#else //DEBUG

struct Log {
  template <typename T>
  Log & operator<<(const T &) {return *this;}
} clog;
#define endl 0

#endif //DEBUG

////////////////////////////////////////////////////////////////////
//
//

struct MutableRow {
  byte * data;
  int pitch;
  MutableRow() : data(0) {}
  MutableRow(byte * d, int p) : data(d), pitch(p) {}
};

struct ConstRow {
  const byte * data;
  int pitch;
  ConstRow() : data(0) {}
  ConstRow(MutableRow o) : data(o.data), pitch(o.pitch) {}
  ConstRow(byte * d, int p) : data(d), pitch(p) {}
};

class VideoFramePtr
{
  AVS_VideoFrame * ptr;
  VideoFramePtr(const VideoFramePtr &);
  void operator= (const VideoFramePtr &);
public:
  VideoFramePtr(AVS_VideoFrame * p = 0) : ptr(p) {}
  void operator= (AVS_VideoFrame * p) {assert(ptr == 0); ptr = p;}
  operator AVS_VideoFrame * () {return ptr;}
  AVS_VideoFrame * release() {AVS_VideoFrame * tmp = ptr; ptr = 0; return tmp;}
  ~VideoFramePtr() {if (ptr) avs_release_video_frame(ptr);}
};

////////////////////////////////////////////////////////////////////
//
//

enum VideoFormat {VF_YV12, VF_YUY2};
enum FieldOrder {FO_Unknown, FO_TFF, FO_BFF};
struct Filter;

struct FieldDiff
{
  int num;
  float diff;
  int   what;
  void clear() {num = -9; diff = 1.0; what = 2;}
  FieldDiff() {clear();}
  FieldDiff(int n, float d, int w) : num(n), diff(d), what(w) {}
};

enum DeinterlaceMethod {WEAVE, USE_EVEN, USE_ODD};

struct FrameInfo {
  // -2: use none, -1: use prev, 0: use cur, 1: use next
  int even;   // even field
  int odd;    // odd  field
  DeinterlaceMethod method;
  FrameInfo() : even(-2), odd(-2) {}
  FrameInfo(int e, int o, DeinterlaceMethod m)
    : even(e), odd(o), method(m) {}
};

//
//
//

struct Analysis
{
  float thres1;
  float thres2;
  float vals[3];
  int last_set;
  int ready;
  int num; // frame number the next adv will give

  void reset(int num);
  FieldDiff adv(float val); // delay of 2
};

struct FieldDiffCache
{
  Filter * bptr;
  static const int SIZE = 16;
  static const int MASK = 15;
  FieldDiff data[SIZE];
  FieldDiff get(int fn) {
    FieldDiff tmp = data[fn & MASK];
    if (tmp.num == fn) return tmp;
    adv_to(fn);
    return data[fn & MASK];
  }
  int diff(int d1, int d2) {
    if (d2 < d1) swap(d1,d2);
    int mx = 0;
    for (; d1 < d2; ++d1)
      mx = max(mx, get(d1).what);
    return mx;
  }
  void clear();
  void adv_to(int fn);
};

struct Filter
{
  bool always_weave;
  bool aggressive;
  
  FieldOrder field_order;
  
  VideoFormat video_format;
  
  float ratio;
  float round_div;

  size_t width;
  size_t row_size, c_row_size;
  size_t height, c_height;
 
  AVS_Clip * child;
  AVS_ScriptEnvironment * env;
  const AVS_VideoInfo * dest_vi;

  int prev_src_fn;
  int cur_src_fn;
  int prev_fn;
  Analysis anal;
  
  float accum;
  int num_src_frames;
  
  FieldDiffCache cache;
  
  void reset(int num);
  
  float compare_fields(ConstRow top, ConstRow bottom);
  float compare(int num);
  int diff(int d1, int d2) {return cache.diff(cur_src_fn + d1, cur_src_fn + d2);}
  bool same_group(int x, int y) {return diff(x,y) <= 0;}
  bool similar_group(int x, int y) {return diff(x,y) <= 1;}
  bool diff_group(int x, int y) {return diff(x,y) == 2;}

  bool chose_next_frame();
  FrameInfo get_frame_info();
  AVS_VideoFrame * render(FrameInfo fi);
  
  static AVS_VideoFrame * get_frame(AVS_FilterInfo *, int num);
};

/////////////////////////////////////////////////////////////////////
//
//
//

#define ERR(str) do {ret = avs_new_value_error(str); goto error;} while (0)
#define SET_I(what, min, max) do {\
    tmp = avs_array_elt(args, i++);\
    if (avs_defined(tmp)) {\
      int v = avs_as_int(tmp);\
      if (v < min || v > max)\
        ERR(#what " must be in the range " #min "-" #max ".");\
      f->what = v;}} while (0)
#define SET_BYTE(what) SET_I(what, 0, 255)
#define SET_UINT(what) SET_I(what, 0, 32500)

AVS_Value smart_decimate(AVS_ScriptEnvironment * env, AVS_Value args, void * )
{
  AVS_Value ret;
  AVS_Value tmp;
  AVS_FilterInfo * fi;
  Filter * f = 0;
  int numr, denm;
  AVS_Clip * clip = avs_new_c_filter(env, &fi, args.d.array[0], 1);
  
  VideoFramePtr frame(avs_get_frame(fi->child, 0));
  
  f = new Filter;
  
  if (avs_is_yuy2(&fi->vi))
    f->video_format = VF_YUY2;
  else if (avs_is_yv12(&fi->vi))
    f->video_format = VF_YV12;
  else
    ERR("Video must be YUY2 or YV12");
  
  if (!avs_is_field_based(&fi->vi)) ERR("Use SeparateFieldsFirst.");
    
  fi->user_data = f;
  fi->get_frame = Filter::get_frame;

  f->prev_fn = -9;  
  f->cache.bptr = f;
  
  numr = avs_as_int(avs_array_elt(args, 1));
  denm = avs_as_int(avs_array_elt(args, 2));
  tmp = avs_array_elt(args, 3);
  f->always_weave = avs_defined(tmp) ? avs_as_bool(tmp) : false;
  tmp = avs_array_elt(args, 4);
  f->aggressive   = avs_defined(tmp) ? avs_as_bool(tmp) : true;
  
  f->child   = fi->child;
  f->env     = fi->env;
  f->dest_vi = &fi->vi; 

  f->num_src_frames = fi->vi.num_frames;

  f->width = fi->vi.width;
  
  f->row_size = avs_get_row_size(frame);  
  f->height   = avs_get_height(frame);
  
  f->c_row_size = avs_get_row_size_p(frame, AVS_PLANAR_U);
  f->c_height   = avs_get_height_p  (frame, AVS_PLANAR_U);
  
  if (avs_get_parity(fi->child, 0)) {
    f->field_order = FO_TFF;
    clog << "Field Order Is TFF" << endl;
  } else {
    f->field_order = FO_BFF;
    clog << "Field Order Is BFF" << endl;
  }
  
  f->ratio = (float)denm/(float)numr;

  fi->vi.height *= 2;
  fi->vi.num_frames = (int)(fi->vi.num_frames/f->ratio);
  avs_set_fps(&fi->vi, fi->vi.fps_numerator * numr, 
                       fi->vi.fps_denominator * denm);
  
  ret = avs_new_value_clip(clip);
  avs_release_clip(clip); clip = 0;
  return ret;
error:
  delete f;
  avs_release_clip(clip); clip = 0;
  return ret;
}

const char * avisynth_c_plugin_init(AVS_ScriptEnvironment * env)
{
  avs_add_function(env, "SmartDecimate", "cii[weave]b[aggressive]b", smart_decimate, 0);
  return "";
}

//////////////////////////////////////////////////////////////////////
//
//
//

#ifdef DEBUG

ostream & operator<< (ostream & o, const FieldDiff & f)
{
  o << f.what << ' ' << f.diff;
  return o;
}

ostream & operator<< (ostream & o, const FrameInfo & finf)
{
  o << '[' << finf.even;
  if (finf.method == USE_EVEN) o << '*';
  o << ", " << finf.odd;
  if (finf.method == USE_ODD)  o << '*';
  o << ']';
  return o;
}

#endif

AVS_VideoFrame * Filter::get_frame(AVS_FilterInfo * fi, int num)
{
  Filter * f = (Filter *)fi->user_data;
  assert(num >= 0);
  assert(num < f->dest_vi->num_frames);
  if (f->prev_fn + 1 != num) {
    f->reset(num);
  } else {
    f->chose_next_frame();
    for (int i = f->prev_src_fn; i < f->cur_src_fn; ++i)
      clog << "Diff " << i << ": " << f->cache.get(i) << endl;
  }
  FrameInfo inf = f->get_frame_info();
  clog << "FRAME " << num << "(" << f->cur_src_fn << ") = " << inf << endl;
  f->prev_fn = num;
  return f->render(inf);
}

void Filter::reset(int num)
{
  clog << "Reseting to " << num << endl;
  cache.clear();
  float ideal = num*ratio;
  cur_src_fn = (int)ideal;
  accum = ideal - cur_src_fn;
  round_div = 1.0;
  anal.reset(cur_src_fn);
  anal.adv(compare(anal.num + 2));
  anal.adv(compare(anal.num + 2));
}

float Filter::compare(int num)
{
  if (num < 0 || num >= num_src_frames)
    return 1.0;
  AVS_VideoFrame * cur = avs_get_frame(child, num);
  int cur_o = avs_get_parity(child, num);
  AVS_VideoFrame * next = avs_get_frame(child, num+1);
  int next_o = avs_get_parity(child, num+1);
  assert(cur_o != next_o);
  ConstRow cur_d, next_d;
  cur_d.data  = avs_get_read_ptr(cur);
  cur_d.pitch = avs_get_pitch(cur);
  next_d.data = avs_get_read_ptr(next);
  next_d.pitch = avs_get_pitch(next);
  float res;
  if (cur_o)
    res = compare_fields(cur_d, next_d);
  else
    res = compare_fields(next_d, cur_d);
  avs_release_video_frame(cur);
  avs_release_video_frame(next);
  return res;
}

void FieldDiffCache::clear()
{
  for (int i = 0; i != SIZE; ++i)
    data[i].clear();
}

void FieldDiffCache::adv_to(int fn)
{
  assert(bptr->anal.num <= fn);
  while (bptr->anal.num <= fn)
  {
    FieldDiff fd = bptr->anal.adv(bptr->compare(bptr->anal.num + 2));
    data[fd.num & MASK] = fd;
  }
}

//////////////////////////////////////////////////////////////////////
//
// Compare
//

static const int VSCALE = 2;
static const int HSCALE = 3;

REGPARM void subsample_line_yuy2_2(const byte * x, short * o, size_t row_size)
{
  const byte * stop = x + row_size;
  *o = x[0] + x[2];
  x += 4;
  ++o;
  while (x < stop) {
    *o = x[-2] + x[0] + x[2];
    x += 4;
    ++o;
  }
}

REGPARM void subsample_line_yuy2_4(const byte * x, short * o, size_t row_size)
{
  const byte * stop = x + row_size;
  *o = x[0] + x[2] + x[4] + x[8];
  x += 8;
  ++o;
  while (x < stop) {
    *o = x[-6] + x[-4] + x[-2] + x[0] + x[2] + x[4] + x[6];
    x += 8;
    ++o;
  }
}

REGPARM void subsample_line_yv12_2(const byte * x, short * o, size_t row_size)
{
  const byte * stop = x + row_size;
  *o = x[0] + x[1];
  x += 2;
  ++o;
  while (x < stop) {
    *o = x[-1] + x[0] + x[1];
    x += 2;
    ++o;
  }
}

REGPARM void subsample_line_yv12_4(const byte * x, short * o, size_t row_size)
{
  const byte * stop = x + row_size;
  *o = x[0] + x[1] + x[2] + x[3];
  x += 4;
  ++o;
  while (x < stop) {
    *o = x[-3] + x[-2] + x[-1] + x[0] + x[1] + x[2] + x[3];
    x += 4;
    ++o;
  }
}

float Filter::compare_fields(ConstRow top, ConstRow bottom)
{
  REGPARM void (* subsample_line)(const byte * x, short * o, size_t row_size);
  if (video_format == VF_YUY2) {
    if (VSCALE == 2)  subsample_line = subsample_line_yuy2_2;
    else              subsample_line = subsample_line_yuy2_4;
  } else {
    if (VSCALE == 2)  subsample_line = subsample_line_yv12_2;
    else              subsample_line = subsample_line_yv12_4;}  

  int line = 0;
  int i;
  const byte * s_l[2] = {top.data, bottom.data};
  int pitch[2] = {top.pitch, bottom.pitch};
  int s_height = height*2/HSCALE - 1;
  int s_width  = width/VSCALE;
  short ad[s_width],bd[s_width],cd[s_width],dd[s_width],ed[s_width];
  short * a = ad, * b = bd, * c = cd, * d = dd, * e = ed;
  subsample_line(top.data, d, row_size);
  subsample_line(bottom.data, e, row_size);
  float accum = 0;
  while (line < s_height)
  {
    swap(a,d);
    swap(b,e);
    i = line & 1; // 0 if even, 1 if odd
    s_l[i] += pitch[i]; subsample_line(s_l[i],c, row_size);
    s_l[i] += pitch[i]; subsample_line(s_l[i],e, row_size);
    i = i ^ 1; // 1 if even, 0 if odd
    s_l[i] += pitch[i]; subsample_line(s_l[i],d, row_size);

    // now compare lines
    for (int x = 0; x < s_width; ++x) 
    {
      int v = a[x] + 4*c[x] + e[x] - 3*(b[x] + d[x]);
      v *= v;
      accum += v;
    }
    ++line;
  }
  static const float scale = 255* 6 * (2*VSCALE - 1);
  accum /= s_width*s_height*scale*scale;
  return accum;
}

//////////////////////////////////////////////////////////////////////
//
// Analysis
//

static const float THRES1_DEF = 0.00010;
static const float THRES2_DEF = 0.00025;

static const int MAX_LAST_SET = 20;
static const float T1_ABS = 0.0025, T1_REL = 0.65, T1_DIFF = 0.07;
static const float T1_T1 = 0.05, T1_T2 = 0.15;
static const float T2_ABS = 0.0040, T2_REL = 0.92, T2_DIFF = 0.10;
static const float T2_T2 = 0.03, T2_T1 = -0.10;

void Analysis::reset(int n)
{
  thres1 = THRES1_DEF;
  thres2 = THRES2_DEF;
  last_set = -1;
  num = n - 2 - 2;
  ready = 0;
}

FieldDiff Analysis::adv(float val)
{
  ++num;
  if (ready < 2) {

    vals[ready] = val;
    ready++;
    return FieldDiff();

  } else {

    vals[2] = val;
    if (vals[0] < vals[1] && vals[2] < vals[1]) {
      float p = vals[0] / vals[1];
      float n = vals[2] / vals[1];
      //cerr << " " << vals[0] << " " << p 
      //     << " " << abs(p - n)
      //     << " " << n << " " << vals[2] << "\n";
      if (vals[0] < T1_ABS && vals[2] < T1_ABS &&
          p < T1_REL && n < T1_REL && abs(p - n) < T1_DIFF) 
      {
        //cerr << " T1\n";
        last_set = 0;
        thres1 = (max(p,n) + T1_T1)*vals[1];
        thres2 = (max(p,n) + T1_T2)*vals[1];
      } 
      else if (vals[0] < T2_ABS && vals[2] < T2_ABS &&
               p < T2_REL && n < T2_REL && abs(p - n) < T2_DIFF) 
      {
        //cerr << " T2\n";
        last_set = 0;
        thres2 = (max(p,n) + T2_T2)*vals[1];
        if (thres1 > thres2) thres1 = (min(p,n) + T2_T1)*vals[1];
      }
    }
    if (last_set > MAX_LAST_SET) 
    {
      //cerr << " RESET\n";
      thres1 = THRES1_DEF;
      thres2 = THRES2_DEF;
      last_set = -1;
    }
    if (last_set >= 0)
      last_set ++;
    int what;
    if (vals[0] < thres1)       what = 0;
    else if (vals[0] < thres2)  what = 1;
    else                        what = 2;
    //cerr << " " << vals[0] << " = " << what << '\n';
    FieldDiff res(num-1, vals[0], what);
    vals[0] = vals[1];
    vals[1] = vals[2];
    return res;
  }
}

//////////////////////////////////////////////////////////////////////
//
// Find Best Frame
//

static const float DELTA = 0.001;

bool Filter::chose_next_frame()
{
  prev_src_fn = cur_src_fn;
  float ideal = prev_src_fn + accum + ratio;
  int     cur = (int)ideal;
  assert(cur < num_src_frames);
  cur_src_fn = cur;
  int p = prev_src_fn - cur;  // note p does not nessary have to equal -1
  float frac  = ideal - cur;
  enum Chosen {Either, Cur, Next};
  int chosen = Cur;

  if (cur == num_src_frames - 1)                          chosen = Cur;
  else if (same_group(0, 1))                              chosen = Either;
  else if (same_group(p, 0))                              chosen = Next;
  else if (similar_group(p,0) && similar_group(-1,0)) {
    if (similar_group(0, 1) && !same_group(1, 2))         chosen = Either;
    else if (diff_group(0, 1))                            chosen = Next;
  } else if (diff_group(p, 0) && diff_group(-1, 0)
             && diff_group(1,2))                          chosen = Either;

  if (chosen == Either) {
    cur_src_fn = frac < round_div ? cur : cur + 1;
  } else if (chosen == Cur) {
    if (round_div < frac) round_div = frac + DELTA;
    cur_src_fn = cur;
  } else if (chosen == Next) {
    if (round_div > frac) round_div = frac - DELTA;
    cur_src_fn = cur + 1;
  }

  accum = ideal - cur_src_fn;
  return true;
}

FrameInfo Filter::get_frame_info()
{
  int matching = is_even(cur_src_fn) ? 1 : -1;
  int other    = -matching;

  enum UseF {Matching, Other, Neither} use_f;
  bool weave = true;
  
  if      (same_group(0, matching))   use_f = Matching;
  else if (same_group(0, other))      use_f = Other;
  else if (aggressive) {
    weave = false;
    if (similar_group(0, matching))   use_f = Matching;
    else if (similar_group(0, other)) use_f = Other;
    else                              use_f = Matching;
  } else {
    weave = false;
    int other_n = -(matching + 1);
    if (!same_group(other, other_n))  use_f = Matching;
    else                              use_f = Neither;
  }
  
  if (field_order == FO_TFF) {
    if (is_even(cur_src_fn))
      return FrameInfo(0, use_f == Matching ? 1 : use_f == Other ? -1 : -2,
                       weave ? WEAVE : USE_EVEN);
    else
      return FrameInfo(use_f == Matching ? -1 : use_f == Other ?  1 : -2, 0,
                       weave ? WEAVE : USE_ODD);
  } else {
    if (is_even(cur_src_fn))
      return FrameInfo(use_f == Matching ? 1 : use_f == Other ? -1 : -2, 0,
                       weave ? WEAVE : USE_ODD);
    else
      return FrameInfo(0, use_f == Matching ? -1 : use_f == Other ?  1 : -2,
                       weave ? WEAVE : USE_EVEN);
  }
}
    
//////////////////////////////////////////////////////////////////////
//
// Render
//

struct MutableRawFrame
{
  VideoFramePtr frame;
  MutableRow y;
  MutableRow chroma[2];
  void init(AVS_VideoFrame * frame);
};

struct ConstRawFrame
{
  VideoFramePtr frame;
  ConstRow y;
  ConstRow chroma[2];
  void init(AVS_VideoFrame * frame);
};

void bob_even(ConstRow src, MutableRow dest, size_t row_size, size_t height);
void bob_odd(ConstRow src, MutableRow dest, size_t row_size, size_t height);
void bob_chroma(const ConstRawFrame & src, const MutableRawFrame & dest,
                size_t row_size, size_t height);
void weave(ConstRow even, ConstRow odd, MutableRow dest,
           size_t row_size, size_t height);
void weave_chroma(const ConstRawFrame & even, const ConstRawFrame & odd,
                  const MutableRawFrame & dest, 
                  size_t row_size, size_t height);

AVS_VideoFrame * Filter::render(FrameInfo finf)
{
  MutableRawFrame dest;
  dest.init(avs_new_video_frame(env, dest_vi, 8));
  finf.even = finf.even == -2 ? -1 : cur_src_fn + finf.even;
  finf.odd  = finf.odd  == -2 ? -1 : cur_src_fn + finf.odd;
  
  DeinterlaceMethod dim = finf.method;

  if (always_weave && finf.even != -1 && finf.odd != -1)
    dim = WEAVE;

  if (dim == WEAVE) {

    ConstRawFrame even, odd;
    even.init(avs_get_frame(child, finf.even));
    odd.init(avs_get_frame(child, finf.odd));
    
    weave(even.y, odd.y, dest.y, row_size, height);
    
    if (video_format == VF_YV12)
      weave_chroma(even, odd, dest, c_row_size, c_height);
  
  } else {

    ConstRawFrame src;
    if (dim == USE_EVEN) {
      src.init(avs_get_frame(child, finf.even));
      bob_even(src.y, dest.y, row_size, height);
    } else {
      src.init(avs_get_frame(child, finf.odd));
      bob_odd(src.y, dest.y, row_size, height);
    }
    if (video_format == VF_YV12)
      bob_chroma(src, dest, c_row_size, c_height);
  }

  return dest.frame.release();
}

void MutableRawFrame::init(AVS_VideoFrame * f)
{
  frame = f;
  y.data  = avs_get_write_ptr(frame);
  y.pitch = avs_get_pitch(frame);
  chroma[0].data = avs_get_write_ptr_p(frame, AVS_PLANAR_U);
  chroma[0].pitch = avs_get_pitch_p(frame, AVS_PLANAR_U);
  chroma[1].data = avs_get_write_ptr_p(frame, AVS_PLANAR_V);
  chroma[1].pitch = avs_get_pitch_p(frame, AVS_PLANAR_V);
}

void ConstRawFrame::init(AVS_VideoFrame * f)
{
  frame = f;
  y.data  = avs_get_read_ptr(frame);
  y.pitch = avs_get_pitch(frame);
  chroma[0].data = avs_get_read_ptr_p(frame, AVS_PLANAR_U);
  chroma[0].pitch = avs_get_pitch_p(frame, AVS_PLANAR_U);
  chroma[1].data = avs_get_read_ptr_p(frame, AVS_PLANAR_V);
  chroma[1].pitch = avs_get_pitch_p(frame, AVS_PLANAR_V);
}

void bob_even(ConstRow src, MutableRow dest, size_t row_size, size_t height)
{
  byte * d0 = dest.data;
  byte * d1 = d0 + dest.pitch;
  const byte * s0 = src.data; 
  const byte * s1 = s0 + src.pitch;
  for (size_t y = 0; y != height - 2; ++y) {
    for (size_t x = 0; x != row_size; ++x) {
      d0[x] = s0[x];
      d1[x] = (s0[x] + s1[x])/2;
    }
    d0 += 2*dest.pitch; d1 += 2*dest.pitch;
    s0 = s1; s1 += src.pitch;
  }
  memcpy(d0, s0, row_size);
  memcpy(d1, s0, row_size); 
}

void bob_odd(ConstRow src, MutableRow dest, size_t row_size, size_t height)
{
  byte * d0 = dest.data;
  byte * d1 = d0 + dest.pitch;
  const byte * s0 = 0; 
  const byte * s1 = src.data;
  memcpy(d0, s1, row_size);
  memcpy(d1, s1, row_size);
  for (size_t y = 1; y != height; ++y) {
    d0 += 2*dest.pitch; d1 += 2*dest.pitch;
    s0 = s1; s1 += src.pitch;
    for (size_t x = 0; x != row_size; ++x) {
      d0[x] = (s0[x] + s1[x])/2;
      d1[x] = s1[x];
    }
  }
}

void bob_chroma(const ConstRawFrame & src, const MutableRawFrame & dest,
                size_t row_size, size_t height)
{
  for (size_t i = 0; i != 2; ++i) {
    byte * d       = dest.chroma[i].data;
    int    d_pitch = dest.chroma[i].pitch;
    const byte * s = src.chroma[i].data;
    int s_pitch    = src.chroma[i].pitch;
    for (size_t y = 0; y != height; ++y) {
      memcpy(d, s, row_size); d += d_pitch;
      memcpy(d, s, row_size); d += d_pitch;
      s += s_pitch;
    }
  }
}


void weave(ConstRow even, ConstRow odd, MutableRow dest,
           size_t row_size, size_t height)
{
  byte * dl = dest.data;
  const byte * el = even.data; const byte * ol = odd.data;
  for (size_t y = 0; y != height; ++y) {
    memcpy(dl, el, row_size); dl += dest.pitch; el += even.pitch;
    memcpy(dl, ol, row_size); dl += dest.pitch; ol += even.pitch;
  }
}

void weave_chroma(const ConstRawFrame & even, const ConstRawFrame & odd,
                  const MutableRawFrame & dest, 
                  size_t row_size, size_t height)
{
  for (int i = 0; i != 2; ++i) {
    int    d_pitch = dest.chroma[i].pitch;
    byte * d0 = dest.chroma[i].data; byte * d1 = d0 + d_pitch;
    const byte * el = even.chroma[i].data; int e_pitch = even.chroma[i].pitch;
    const byte * ol = odd.chroma[i].data;  int o_pitch = odd.chroma[i].pitch;
    for (size_t y = 0; y != height; ++y) {
      for (size_t x = 0; x != row_size; ++x) {
        d0[x] = (el[x] + ol[x])/2;
        d1[x] = d0[x];
      }
      d0 += 2*d_pitch; d1 += 2*d_pitch;
      el += e_pitch; ol += o_pitch;
    }
  }
}

